/*
	Copyright Andrew Varner 2011.07.22

	This file is part of rohttpfs.

	rohttpfs is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License version 3 as
	published by the Free Software Foundation.

	rohttpfs is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with rohttpfs.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef ROHTTPFS_FILE_H
#define ROHTTPFS_FILE_H

#include <string>
#include <time.h>
#include <errno.h>
#include <list>
#include <vector>


class rohttpfs;


struct file_block_element
{
	char * data;
	std::list<file_block_element *>::iterator iter_list;
};



class file
{
	private:
		rohttpfs * server;
		uint64_t /*TODO: use fastest*/ open_count;
		time_t age;
		time_t idle;
		
		std::string path;
		int type; //TODO: type should be an enum
		size_t size;
		bool size_unknown;
		time_t mtime;
		//TODO: other file meta data
		
		//TODO: map of blocks
		std::vector<file_block_element> data_cache;
		
		ssize_t read_direct(char * buf, size_t requested_len, off_t offset);
	public:
		file(const std::string path1, rohttpfs * server1);
		void open();
		void close();
		
		std::string get_path();
		int get_type();
		size_t get_size();
		time_t get_mtime();
		int read(char * buf, size_t requested_len_u, off_t offset);
		uint64_t get_open_count();
		~file();
	private:
		void refresh_data();
		int fill_buffer(size_t block_num, bool force);
};

#endif

