/*
	Copyright Andrew Varner 2011.07.22

	This file is part of rohttpfs.

	rohttpfs is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License version 3 as
	published by the Free Software Foundation.

	rohttpfs is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with rohttpfs.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "main.h"

#include <string.h>
#include <unistd.h>
#include <netinet/in.h>

#include "rohttpfs.h"


int rohttpfs::get_file_info(/*const struct sockaddr_in * remote_host, const char * remote_host_name,*/ const char * path, uint64_t * size, time_t * mtime)
{
	//TODO: make this function a member of file::
	//TODO: make this function use file::read to get its data (chicken/egg problem)
	
	/*
	TODO: this should be an enum in ::file
	0	does not exist
	1	dir
	2	file
	3	a file/dir, which you will not be able to read (access denied)
	4	access denied, even to the parent directory
	5	I/O error
	*/

	//strip off trailing slashes
	//first, head path.
	//	if size is provided, and it's larger than max_dir_index_size, it's a file
	//	if 404, path does not exist
	//	if 403, return access denied
	//	if 3XX, redirected to "path/", it's a dir
	//	if other error
	//		stat one level up
	//			if access denied, then return access denied
	//			if I/O error, then return I/O error
	//			if it DNE, then path DNE
	//			if it's a file, then path DNE
	//			if it's a dir, then return I/O error
	//get path - first max_dir_index_size+1 bytes only
	//	if it's at least max_dir_index_size+1 bytes, it's a file.
	//	if it is a binary, it's a file.
	//	parse it for links.
	//		if it contains links to items below path, it's a directory.
	//head path/doesnotexist-${RANDOMHEXDIGITS}" ; if a size is not provided, then get path/doesnotexist-${RANDOMHEXDIGITS}"
	//	if exists, and is same size as path, then path is a file
	//	if exists, and is different, then path is a directory
	//	if 404, path is a directory
	//	if any other error, then path is a file

	//TODO:
	//other possibilities:
	//look for unicode/utf-8, not just ascii
	//search for "<title>index of"
	//look up parent index - links to directories end with "/", and the parent directory should already be cached.
	//servers sometimes don't send 206 for automatic indexes
	//servers sometimes don't send file size for automatic indexes, but do for regular files
	
	uint64_t size_dummy = 0;
	if (size == NULL)
		size = &size_dummy;
	time_t mtime_dummy = 0;
	if (mtime == NULL)
		mtime = &mtime_dummy;
	*mtime = 0; //TODO: mtime must be set
	
	const size_t len = strlen(path) + 1;
	char path2[len];
	strncpy(path2, path, len - 1);
	path2[len - 1] = 0;
	for (size_t i = len - 2; (i > 0) && (path2[i] == '/'); i--)
		path2[i] = 0;
	
	int64_t size2;
	int status;
	char redirectbuf[5000];
	status = send_request(/*remote_host, remote_host_name,*/ NULL, 1, redirectbuf, 5000, path2, &size2, 0, -1);
	*size = size2;
	if (strcmp(path, "") == 0)
		return 1;
	if (strcmp(path, "/") == 0)
		return 1;
	if (strcmp(path, "//") == 0)
		return 1;
	if (status == 404)
		return 0;
	else if (status == 403)
		return 4; //TODO: this could be a return 3. add checks.
	else if (status == 200)
	{
		if ((size2 > 0) && (((uint64_t)size2) > max_dir_index_size))
		{
			*size = size2;
			return 2;
		}
	}
	else if (status == 0)
		return 5; //connection timed out
	else
	{
		if (is_this_a_dir_redirect(path, redirectbuf))
			return 1;
		for (size_t i = strlen(path2); (i > 0) && (path2[i] != '/'); i--)
			path2[i] = 0;
		int parent_type = get_file_info(/*remote_host, remote_host_name,*/ path2, NULL, NULL);
		if (parent_type == 2)
			return 0;
		if (parent_type == 1)
			return 5;
		if (parent_type == 3)
			return 4;
		return parent_type;
	}
	
	int s = 0;
	status = send_request(/*remote_host, remote_host_name,*/ &s, 0, NULL, 0, path2, &size2, 0, max_dir_index_size + 1);
	if ((status != 200) && (status != 206))
	{
		close(s);
		if (status == 404)
			return 0;
		if (status == 403)
			return 4;
		return 5;
	}
	char recv_buf[max_dir_index_size + 2];
	ssize_t data_size = read(s, recv_buf, max_dir_index_size + 1);
	close(s);
	if (data_size < 0)
		return 5;
	if (((size_t)data_size) > max_dir_index_size)
		return 2;
	if (!is_this_ascii(recv_buf, data_size))
		return 2;
	recv_buf[data_size + 1] = 0;
	//const char * basepath = NULL; //TODO: fix that
	if (are_there_sublinks(recv_buf/*, path2, basepath*/))
		return 1;
	return 2; //TODO: add the last part of the algorithm
}

