/*
	Copyright Andrew Varner 2011.07.22

	This file is part of rohttpfs.

	rohttpfs is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License version 3 as
	published by the Free Software Foundation.

	rohttpfs is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with rohttpfs.  If not, see <http://www.gnu.org/licenses/>.
*/


#include "main.h"

#include <fuse.h>
#include <unistd.h>
#include <string.h>
#include <stdlib.h>
#include <stdio.h>
#include <errno.h>
#include <netinet/in.h>

#include "rohttpfs.h"
#include "file.h"



int uid = -1;
int gid = -1;



rohttpfs * server;




/*
TODO:

add some command line options
	base URL
	remote host
	connection timeout time
	control max_dir_index_size
	user agent
*/






/*
important functions:
int(* 	getattr )(const char *, struct stat *)
#int(* 	open )(const char *, struct fuse_file_info *)
int(* 	read )(const char *, char *, size_t, off_t, struct fuse_file_info *)
#int(* 	opendir )(const char *, struct fuse_file_info *)
int(* 	readdir )(const char *, void *, fuse_fill_dir_t, off_t, struct fuse_file_info *)
#int(* 	fgetattr )(const char *, struct stat *, struct fuse_file_info *)
unsigned int 	flag_reserved: 31
#int(* 	release )(const char *, struct fuse_file_info *)
#int(* 	releasedir )(const char *, struct fuse_file_info *)
#int(* 	access )(const char *, int)
unsigned int 	flag_nullpath_ok: 1

unimportant stub functions:
int(* 	readlink )(const char *, char *, size_t)
int(* 	getxattr )(const char *, const char *, char *, size_t)
int(* 	listxattr )(const char *, char *, size_t)
#void *(* 	init )(struct fuse_conn_info *conn)
#void(* 	destroy )(void *)
int(* 	statfs )(const char *, struct statvfs *)
int(* 	bmap )(const char *, size_t blocksize, uint64_t *idx)

functions that are unimportant stubs because the filesystem is read only:
int(* 	mknod )(const char *, mode_t, dev_t)
int(* 	mkdir )(const char *, mode_t)
int(* 	unlink )(const char *)
int(* 	rmdir )(const char *)
int(* 	symlink )(const char *, const char *)
int(* 	rename )(const char *, const char *)
int(* 	link )(const char *, const char *)
int(* 	chmod )(const char *, mode_t)
int(* 	chown )(const char *, uid_t, gid_t)
int(* 	truncate )(const char *, off_t)
int(* 	write )(const char *, const char *, size_t, off_t, struct fuse_file_info *)
int(* 	setxattr )(const char *, const char *, const char *, size_t, int)
int(* 	removexattr )(const char *, const char *)
#int(* 	create )(const char *, mode_t, struct fuse_file_info *)
#int(* 	ftruncate )(const char *, off_t, struct fuse_file_info *)
#int(* 	lock )(const char *, struct fuse_file_info *, int cmd, struct flock *)
int(* 	utime )(const char *, struct utimbuf *)
int(* 	utimens )(const char *, const struct timespec tv[2])
#int(* 	flush )(const char *, struct fuse_file_info *)
#int(* 	fsync )(const char *, int, struct fuse_file_info *)
#int(* 	fsyncdir )(const char *, int, struct fuse_file_info *)

status unknown:
int(* 	ioctl )(const char *, int cmd, void *arg, struct fuse_file_info *, unsigned int flags, void *data)
int(* 	poll )(const char *, struct fuse_file_info *, struct fuse_pollhandle *ph, unsigned *reventsp)
*/





int fuse_getattr(const char * path, struct stat * buf)
{
	memset(buf, 0, sizeof(buf));
	
	file * f = server->open(path);
	int type = f->get_type();
	size_t size = f->get_size();
	time_t mtime = f->get_mtime();
	f->close();
	f = NULL;

	switch (type)
	{
		case 0:
		{
			return -ENOENT;
		}
		case 1:
		{
			buf->st_mode = 0555 | S_IFDIR;
			size = 0;
			break;
		}
		case 2:
		{
			buf->st_mode = 0444 | S_IFREG;
			break;
		}
		case 3:
		case 4:
		{
			if (strcmp(path, "/") == 0)
				buf->st_mode = 0 | S_IFDIR;
			else
				buf->st_mode = 0 | S_IFREG;
			break;
		}
		case 5:
		{
			return -EIO;
		}
	}
	
	buf->st_dev = 0;
	buf->st_ino = 0;
	buf->st_nlink = 1;
	buf->st_uid = uid;
	buf->st_gid = gid;
	//buf->st_rdev = ?;
	buf->st_size = size;
	//buf->st_blksize = 1048576;
	buf->st_blocks = size / BLOCK_SIZE + ((size % BLOCK_SIZE)!=0);
	buf->st_atime = mtime;
	buf->st_mtime = mtime;
	buf->st_ctime = mtime;
	return 0;
}


int fuse_readdir(const char * path, void * buf, fuse_fill_dir_t filler, off_t, struct fuse_file_info *)
{
	file * f = server->open(path);
	if (f->get_type() != 1)
	{
		fprintf(logfile, "attempt to readdir from a nondir (fuse should have caught this): %s\n", path); //TODO: test if fuse actually catches this with 'ls rohttpfs/some/dir/'
		f->close();
		return -EIO;
	}

	std::string path2(path);
	path2.append("/");
	
	char index[max_dir_index_size + 1];
	ssize_t read_size = f->read(index, max_dir_index_size, 0);
	f->close();
	if (read_size < 0)
	{
		fprintf(logfile, "read error during readdir of: \"%s\": errno=%d\n", path, (int)-read_size);
		if (strcmp(path, "/") == 0)
		{
			filler(buf, ".", NULL, 0);
			filler(buf, "..", NULL, 0);
			return 0;
		}
		else
		{
			return -EIO;
		}
	}
	
	size_t size = read_size;
	index[size] = 0;
	//TODO: should index be allowed to have null characters? I don't care.

	filler(buf, ".", NULL, 0);
	filler(buf, "..", NULL, 0);
	
	//TODO: parse "index" for links
	const char * cursor = index;
	while (true)
	{
		cursor = strcasestr(cursor, "href=");
		if (cursor == NULL)
			return 0; //TODO: what is correct return value?
		cursor += 5;
		const char * link_start;
		const char * link_end;
		if ((*cursor == '\"') || (*cursor == '\''))
		{
			char q = *cursor;
			link_start = cursor + 1;
			link_end = strchr(link_start, q);
			if (link_end == NULL)
				return 0;
		}
		else
		{
			const char * link_start = cursor;
			const char * link_end = link_start;
			for (; (*link_end > ' ') && (*link_end != '>'); link_end++)
				;
		}
		cursor = link_end;
		
		if (*link_start == '/')
			continue;
		if (strncmp(link_start, "./", 2) == 0)
			link_start += 2;
		while ((*(link_end - 1) == '/') && (link_start < link_end))
			link_end--;
		size_t size = link_end - link_start;
		if (size <= 0)
			continue;
		char link[size + 1];
		memcpy(link, link_start, size);
		link[size] = 0;
		bool valid = true;
		for (size_t i = 0; i < size; i++)
		{
			char c = link[i];
			if ((c == '/') || ((c < ' ') && (c != '\n') && (c != '\r')) || (c > '~') || (c < 0))
			{
				valid = false;
				break;
			}
		}
		if (strcmp(link, "..") == 0)
			valid = false;
		if (valid)
		{
			filler(buf, link, NULL, 0);
			fprintf(logfile, "filling dir: %s (in %s)\n", link, path);
		}
	}
}
int fuse_readdir_debug(const char * path, void * buf, fuse_fill_dir_t filler, off_t offset, struct fuse_file_info * info)
{
	fprintf(logfile, "readdir(\"%s\", x)\n", path);
	fflush(logfile);
	int retval = fuse_readdir(path, buf, filler, offset, info);
	fprintf(logfile, "readdir()=%d\n", retval);
	return retval;
}


int fuse_read_debug(const char * fname, char * buf, size_t requested_len, off_t offset, struct fuse_file_info * info);
int fuse_read(const char * fname, char * buf, size_t requested_len, off_t offset, struct fuse_file_info * info);
int fuse_open_debug(const char * fname, struct fuse_file_info * info);
int fuse_open(const char * fname, struct fuse_file_info * info);


int fuse_getattr_debug(const char * path, struct stat * buf)
{
	fprintf(logfile, "getattr(%s, x)\n", path);
	fflush(logfile);
	int retval = fuse_getattr(path, buf);
	fprintf(logfile, "getattr()=%d\n", retval);
	return retval;
}


int fuse_read_debug(const char * fname, char * buf, size_t requested_len, off_t offset, struct fuse_file_info * info)
{
	fprintf(logfile, "read(%s, x, %lu, %lu, x)\n", fname, requested_len, offset);
	fflush(logfile);
	int retval = fuse_read(fname, buf, requested_len, offset, info);
	fprintf(logfile, "read()=%d\n", retval);
	return retval;
}

int fuse_read(const char * fname, char * buf, size_t requested_len_u, off_t offset, struct fuse_file_info *)
{
	file * f = server->open(fname);
	if (f->get_type() == 1)
		return -EISDIR;
	int retval = f->read(buf, requested_len_u, offset);
	f->close();
	return retval;
}

int fuse_open_debug(const char * fname, struct fuse_file_info * info)
{
	fprintf(logfile, "open(%s, x)\n", fname);
	fflush(logfile);
	int retval = fuse_open(fname, info);
	fprintf(logfile, "open()=%d\n", retval);
	return retval;
}

int fuse_open(const char * fname, struct fuse_file_info * info)
{
	file * f = server->open(fname);
	int retval;
	switch (f->get_type())
	{
		case 0:
			retval = -ENOENT;
			break;
		case 1:
			retval = 0; //TODO: this is a directory - test if 0 is correct return value
			break;
		case 2:
			retval = 0;
			break;
		case 3:
		case 4:
			retval = -EACCES;
			break;
		case 5:
			retval = -EIO;
			break;
		default:
			retval = -EIO;
			break;
	}
	if ((info->flags & (O_RDONLY | O_WRONLY | O_RDWR)) != O_RDONLY)
		retval = -EACCES;
	if (retval != 0)
		f->close();
	return retval;
}

int fuse_release(const char * fname, struct fuse_file_info *)
{
	file * f = server->open(fname);
	f->close();
	f->close();
	return 0;
}
int fuse_release_debug(const char * fname, struct fuse_file_info * info)
{
	fprintf(logfile, "release(%s, x)\n", fname);
	fflush(logfile);
	int retval = fuse_release(fname, info);
	fprintf(logfile, "release()=%d\n", retval);
	return retval;
}



FILE * logfile = NULL;


int main(int nargs, char * args[])
{
	int logfd = creat("log", 0644);
	//dup2(logfd, 2);
	
	logfile = stderr; // = fopen("log", "w");
	//server = new rohttpfs("192.168.0.4");
	server = new rohttpfs("avarner.no-ip.org");
	nargs = 0;
	args = NULL;
	uid = geteuid();
	gid = getegid();
	struct fuse_operations fuse_op;
	memset(&fuse_op, 0, sizeof(fuse_op));
		fuse_op.getattr	= &fuse_getattr;
		fuse_op.readdir	= &fuse_readdir;
		fuse_op.open	= &fuse_open;
//		fuse_op.release	= &fuse_release;
		fuse_op.read	= &fuse_read;

	const char * args2[] = {"rohttpfs", "/home/commodorelxiv/Projects/rohttpfs/mountpoint/", "-s", "-d"};
	char * * args3 = new char *[5];
	args3[0] = new char[strlen(args2[0]) + 1];
	args3[1] = new char[strlen(args2[1]) + 1];
	args3[2] = new char[strlen(args2[2]) + 1];
	args3[3] = new char[strlen(args2[3]) + 1];
	args3[4] = NULL;
	strcpy(args3[0], args2[0]);
	strcpy(args3[1], args2[1]);
	strcpy(args3[2], args2[2]);
	strcpy(args3[3], args2[3]);
	int retval = fuse_main(4, args3, &fuse_op, NULL);
	return retval; //TODO: remove this line, if it's ok
	delete server;
	delete[] args[0];
	delete[] args[1];
	delete[] args;
	return retval;
}


//TODO: urlencode/decode all file names

