/*
	Copyright Andrew Varner 2011.07.22

	This file is part of rohttpfs.

	rohttpfs is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License version 3 as
	published by the Free Software Foundation.

	rohttpfs is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with rohttpfs.  If not, see <http://www.gnu.org/licenses/>.
*/


#define _GNU_SOURCE

#include "main.h"

#include <stdint.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>


bool is_this_ascii(const char * data, size_t length)
{
	for (size_t i = 0; i < length; i++)
	{
		char c = data[i];
		if (c >= 127)
			return 0;
		if (c < ' ')
		{
			if ((c == '\n') || (c == '\r') || (c == '\t'))
				continue;
			return 0;
		}
	}
	return 1;
}

bool is_this_a_dir_redirect_d(const char * path, const char * redirected_url)
{
	bool remove_host_name = false;
	if (strncmp(redirected_url, "//", 2) == 0)
	{
		remove_host_name = true;
	}
	else
	{
		const char * first_slash = strchr(redirected_url, '/');
		const char * first_colon = strchr(redirected_url, ':');
		if ((first_colon - redirected_url) < (first_slash - redirected_url))
		{
			redirected_url = first_slash;
			remove_host_name = true;
		}
	}
	if (remove_host_name)
	{
		while (redirected_url[0] == '/')
			redirected_url++;
		redirected_url = strchrnul(redirected_url, '/');
	}
	//TODO: the following needs the correct basepath (not the hardcoded "rohttpfs")
	const char * base_path = "/rohttpfs";
	if (strncmp(redirected_url, base_path, strlen(base_path)) != 0)
		return false;
	redirected_url += strlen(base_path);
	size_t path_strlen = strlen(path);
	if (strncmp(redirected_url, path, path_strlen) == 0)
	{
		if (strcmp(redirected_url += path_strlen, "/") == 0)
			return true;
		else
			return false;
	}
	else
	{
		return false;
	}
}

bool is_this_a_dir_redirect(const char * path, const char * redirected_url)
{
	//fprintf(logfile, "is_this_a_dir_redirect(\"%s\", \"%s\")\n", path, redirected_url);
	bool retval = is_this_a_dir_redirect_d(path, redirected_url);
	//fprintf(logfile, "is_this_a_dir_redirect()=%d\n", (int)retval);
	return retval;
}

