/*
	Copyright Andrew Varner 2011.07.22

	This file is part of rohttpfs.

	rohttpfs is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License version 3 as
	published by the Free Software Foundation.

	rohttpfs is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with rohttpfs.  If not, see <http://www.gnu.org/licenses/>.
*/


#ifndef ROHTTPFS_ROHTTPFS_H
#define ROHTTPFS_ROHTTPFS_H


#include "file.h"
#include <map>
#include <vector>
#include <list>
#include <netinet/in.h>


class rohttpfs
{
	private:
		struct file_element
		{
			file * f;
			std::list<file_element *>::iterator iter_list;
			std::map<std::string, file_element *>::iterator iter_map;
		};
		
		std::string remote_host_name;
		struct sockaddr_in remote_host;
		
		std::list<file_element *> list_by_age;
		std::map<std::string, file_element *> map_by_name;
		
		std::list<file_block_element *> file_blocks;
		
		char lock;

		bool ping_server();
	public:
		rohttpfs(std::string remote_host_name1);
		file * open(std::string name);
		int send_request(int * s, bool head_only, char * redir_location, size_t redir_location_length, const char * relative_path, int64_t * output_size, size_t range_1, size_t range_2);
		int get_file_info(const char * path, uint64_t * size, time_t * mtime);
		
		void put_new_file_block(file_block_element * f);
		void renew_block(file_block_element * f);
		void remove_block(file_block_element * f);
		
		void renew(std::string name);
		void renew(file_element * f);

		void dump();
};

#endif

